@@ -1,25 +1,24 @@
-;;; forge-gitlab.el --- Gitlab support            -*- lexical-binding: t -*-
+;;; forge-gitlab.el --- Gitlab support  -*- lexical-binding:t -*-
 
-;; Copyright (C) 2018-2022  Jonas Bernoulli
+;; Copyright (C) 2018-2024 Jonas Bernoulli
+
+;; Author: Jonas Bernoulli <emacs.forge@jonas.bernoulli.dev>
+;; Maintainer: Jonas Bernoulli <emacs.forge@jonas.bernoulli.dev>
 
-;; Author: Jonas Bernoulli <jonas@bernoul.li>
-;; Maintainer: Jonas Bernoulli <jonas@bernoul.li>
 ;; SPDX-License-Identifier: GPL-3.0-or-later
 
-;; This file is not part of GNU Emacs.
-
-;; Forge is free software; you can redistribute it and/or modify it
-;; under the terms of the GNU General Public License as published by
-;; the Free Software Foundation; either version 3, or (at your option)
-;; any later version.
+;; This file is free software: you can redistribute it and/or modify
+;; it under the terms of the GNU General Public License as published
+;; by the Free Software Foundation, either version 3 of the License,
+;; or (at your option) any later version.
 ;;
-;; Forge is distributed in the hope that it will be useful, but WITHOUT
-;; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
-;; License for more details.
+;; This file is distributed in the hope that it will be useful,
+;; but WITHOUT ANY WARRANTY; without even the implied warranty of
+;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+;; GNU General Public License for more details.
 ;;
 ;; You should have received a copy of the GNU General Public License
-;; along with Forge.  If not, see http://www.gnu.org/licenses.
+;; along with this file.  If not, see <https://www.gnu.org/licenses/>.
 
 ;;; Code:
 
@@ -48,10 +47,12 @@
 ;;; Pull
 ;;;; Repository
 
-(cl-defmethod forge--pull ((repo forge-gitlab-repository) until)
-  (let ((cb (let ((buf (and (derived-mode-p 'magit-mode)
-                            (current-buffer)))
-                  (dir default-directory)
+(cl-defmethod forge--pull ((repo forge-gitlab-repository)
+                           &optional callback since)
+  (cl-assert (not (and since (forge-get-repository repo nil :tracked?))))
+  (setq forge--mode-line-buffer (current-buffer))
+  (forge--msg repo t nil "Pulling REPO")
+  (let ((cb (let ((buf (current-buffer))
                   (val nil))
               (lambda (cb &optional v)
                 (when v (if val (push v val) (setq val v)))
@@ -67,23 +68,25 @@
                     (forge--fetch-labels repo cb))
                    ((and .issues_enabled
                          (not (assq 'issues val)))
-                    (forge--fetch-issues repo cb until))
+                    (forge--fetch-issues repo cb since))
                    ((and .merge_requests_enabled
                          (not (assq 'pullreqs val)))
-                    (forge--fetch-pullreqs repo cb until))
+                    (forge--fetch-pullreqs repo cb since))
                    (t
                     (forge--msg repo t t   "Pulling REPO")
                     (forge--msg repo t nil "Storing REPO")
-                    (emacsql-with-transaction (forge-db)
+                    (closql-with-transaction (forge-db)
                       (forge--update-repository repo val)
                       (forge--update-assignees  repo .assignees)
                       (forge--update-labels     repo .labels)
                       (dolist (v .issues)   (forge--update-issue repo v))
                       (dolist (v .pullreqs) (forge--update-pullreq repo v))
-                      (oset repo sparse-p nil))
+                      (oset repo condition :tracked))
                     (forge--msg repo t t "Storing REPO")
-                    (unless (oref repo selective-p)
-                      (forge--git-fetch buf dir repo)))))))))
+                    (cond
+                     ((oref repo selective-p))
+                     (callback (funcall callback))
+                     ((forge--maybe-git-fetch repo buf))))))))))
     (funcall cb cb)))
 
 (cl-defmethod forge--fetch-repository ((repo forge-gitlab-repository) callback)
@@ -117,15 +120,15 @@
     (oset repo stars          .star_count)
     (oset repo watchers       .star_count)))
 
-(cl-defmethod forge--split-url-path
-  ((_class (subclass forge-gitlab-repository)) path)
-  (and (string-match "\\`\\(?:~?\\(.+\\)/\\)?\\([^/]+?\\)\\'" path)
-       (list (match-string 1 path)
-             (match-string 2 path))))
+;;;; Topics
+
+(cl-defmethod forge--pull-topic ((repo forge-gitlab-repository) _topic
+                                 &key callback _errorback)
+  (forge--pull repo callback)) ; TODO Pull only the one topic.
 
 ;;;; Issues
 
-(cl-defmethod forge--fetch-issues ((repo forge-gitlab-repository) callback until)
+(cl-defmethod forge--fetch-issues ((repo forge-gitlab-repository) callback since)
   (let ((cb (let (val cur cnt pos)
               (lambda (cb &optional v)
                 (cond
@@ -150,7 +153,8 @@
     (forge--glab-get repo "/projects/:project/issues"
       `((per_page . 100)
         (order_by . "updated_at")
-        (updated_after . ,(forge--topics-until repo until 'issue)))
+        ,@(and-let* ((after (or since (oref repo issues-until))))
+            `((updated_after . ,after))))
       :unpaginate t
       :callback (lambda (value _headers _status _req)
                   (funcall cb cb value)))))
@@ -166,23 +170,27 @@
                   (funcall cb cb)))))
 
 (cl-defmethod forge--update-issue ((repo forge-gitlab-repository) data)
-  (emacsql-with-transaction (forge-db)
+  (closql-with-transaction (forge-db)
     (let-alist data
       (let* ((issue-id (forge--object-id 'forge-issue repo .iid))
              (issue
               (forge-issue
                :id           issue-id
-               :repository   (oref repo id)
+               :their-id     .iid
                :number       .iid
+               :slug         (format "#%s" .iid)
+               :repository   (oref repo id)
+               ;; Gitlab doesn't make a distinction between completed
+               ;; and unplanned issues.  Treat them all as completed.
                :state        (pcase-exhaustive .state
-                               ("closed" 'closed)
+                               ("closed" 'completed)
                                ("opened" 'open))
                :author       .author.username
                :title        .title
                :created      .created_at
                :updated      .updated_at
                ;; `.closed_at' may be nil even though the issues is
-               ;; closed.  In such cases use 1, so that this slots
+               ;; closed.  In such cases use 1, so that this slot
                ;; at least can serve as a boolean.
                :closed       (or .closed_at (and (equal .state "closed") 1))
                :locked-p     .discussion_locked
@@ -192,7 +200,6 @@
         (unless (magit-get-boolean "forge.omitExpensive")
           (forge--set-id-slot repo issue 'assignees .assignees)
           (forge--set-id-slot repo issue 'labels .labels))
-        .body .id ; Silence Emacs 25 byte-compiler.
         (dolist (c .notes)
           (let-alist c
             (let ((post
@@ -204,11 +211,15 @@
                     :created .created_at
                     :updated .updated_at
                     :body    (forge--sanitize-string .body))))
-              (closql-insert (forge-db) post t))))))))
+              (closql-insert (forge-db) post t))))
+        (let ((until (oref repo issues-until)))
+          (when (or (not until) (string> .updated_at until))
+            (oset repo issues-until .updated_at)))
+        issue))))
 
 ;;;; Pullreqs
 
-(cl-defmethod forge--fetch-pullreqs ((repo forge-gitlab-repository) callback until)
+(cl-defmethod forge--fetch-pullreqs ((repo forge-gitlab-repository) callback since)
   (let ((cb (let (val cur cnt pos)
               (lambda (cb &optional v)
                 (cond
@@ -237,7 +248,8 @@
     (forge--glab-get repo "/projects/:project/merge_requests"
       `((per_page . 100)
         (order_by . "updated_at")
-        (updated_after . ,(forge--topics-until repo until 'pullreq)))
+        ,@(and-let* ((after (or since (oref repo pullreqs-until))))
+            `((updated_after . ,after))))
       :unpaginate t
       :callback (lambda (value _headers _status _req)
                   (funcall cb cb value)))))
@@ -275,24 +287,26 @@
   (let-alist (car cur)
     (forge--glab-get repo (format "/projects/%s" .target_project_id) nil
       :errorback (lambda (_err _headers _status _req)
-                   (setf (alist-get 'source_project (car cur)) nil)
+                   (setf (alist-get 'target_project (car cur)) nil)
                    (funcall cb cb))
       :callback (lambda (value _headers _status _req)
                   (setf (alist-get 'target_project (car cur)) value)
                   (funcall cb cb)))))
 
 (cl-defmethod forge--update-pullreq ((repo forge-gitlab-repository) data)
-  (emacsql-with-transaction (forge-db)
+  (closql-with-transaction (forge-db)
     (let-alist data
       (let* ((pullreq-id (forge--object-id 'forge-pullreq repo .iid))
              (pullreq
               (forge-pullreq
                :id           pullreq-id
-               :repository   (oref repo id)
+               :their-id     .iid
                :number       .iid
+               :slug         (format "!%s" .iid)
+               :repository   (oref repo id)
                :state        (pcase-exhaustive .state
                                ("merged" 'merged)
-                               ("closed" 'closed)
+                               ("closed" 'rejected)
                                ("opened" 'open))
                :author       .author.username
                :title        .title
@@ -306,22 +320,25 @@
                                  (and (member .state '("closed" "merged")) 1))
                :merged       (or .merged_at
                                  (and (equal .state "merged") 1))
+               :draft-p      .draft
                :locked-p     .discussion_locked
                :editable-p   .allow_maintainer_to_push
                :cross-repo-p (not (equal .source_project_id
                                          .target_project_id))
                :base-ref     .target_branch
+               :base-rev     .diff_refs.start_sha
                :base-repo    .target_project.path_with_namespace
                :head-ref     .source_branch
+               :head-rev     .diff_refs.head_sha
                :head-user    .source_project.owner.username
                :head-repo    .source_project.path_with_namespace
                :milestone    .milestone.iid
                :body         (forge--sanitize-string .description))))
         (closql-insert (forge-db) pullreq t)
         (unless (magit-get-boolean "forge.omitExpensive")
-          (forge--set-id-slot repo pullreq 'assignees (list .assignee))
+          (forge--set-id-slot repo pullreq 'assignees .assignees)
+          (forge--set-id-slot repo pullreq 'review-requests .reviewers)
           (forge--set-id-slot repo pullreq 'labels .labels))
-        .body .id ; Silence Emacs 25 byte-compiler.
         (dolist (c .notes)
           (let-alist c
             (let ((post
@@ -333,7 +350,11 @@
                     :created .created_at
                     :updated .updated_at
                     :body    (forge--sanitize-string .body))))
-              (closql-insert (forge-db) post t))))))))
+              (closql-insert (forge-db) post t))))
+        (let ((until (oref repo pullreqs-until)))
+          (when (or (not until) (string> .updated_at until))
+            (oset repo pullreqs-until .updated_at)))
+        pullreq))))
 
 ;;;; Other
 
@@ -366,7 +387,7 @@
 (cl-defmethod forge--fetch-forks ((repo forge-gitlab-repository) callback)
   (forge--glab-get repo "/projects/:project/forks"
     '((per_page . 100)
-      (simple . "true"))
+      (simple . t))
     :unpaginate t
     :callback (lambda (value _headers _status _req)
                 (funcall callback callback (cons 'forks value)))))
@@ -427,24 +448,6 @@
 
 ;;; Mutations
 
-(cl-defmethod forge--submit-create-pullreq ((_ forge-gitlab-repository) base-repo)
-  (let-alist (forge--topic-parse-buffer)
-    (pcase-let* ((`(,base-remote . ,base-branch)
-                  (magit-split-branch-name forge--buffer-base-branch))
-                 (`(,head-remote . ,head-branch)
-                  (magit-split-branch-name forge--buffer-head-branch))
-                 (head-repo (forge-get-repository 'stub head-remote)))
-      (forge--glab-post head-repo "/projects/:project/merge_requests"
-        `(,@(and (not (equal head-remote base-remote))
-                 `((target_project_id . ,(oref base-repo forge-id))))
-          (target_branch . ,base-branch)
-          (source_branch . ,head-branch)
-          (title         . , .title)
-          (description   . , .body)
-          (allow_collaboration . t))
-        :callback  (forge--post-submit-callback)
-        :errorback (forge--post-submit-errorback)))))
-
 (cl-defmethod forge--submit-create-issue ((_ forge-gitlab-repository) repo)
   (let-alist (forge--topic-parse-buffer)
     (forge--glab-post repo "/projects/:project/issues"
@@ -453,6 +456,28 @@
       :callback  (forge--post-submit-callback)
       :errorback (forge--post-submit-errorback))))
 
+(cl-defmethod forge--submit-create-pullreq ((_ forge-gitlab-repository) base-repo)
+  (let-alist (forge--topic-parse-buffer)
+    (pcase-let* ((`(,base-remote . ,base-branch)
+                  (magit-split-branch-name forge--buffer-base-branch))
+                 (`(,head-remote . ,head-branch)
+                  (magit-split-branch-name forge--buffer-head-branch))
+                 (head-repo (forge-get-repository :stub head-remote)))
+      (forge--glab-post head-repo "/projects/:project/merge_requests"
+        `((title . ,(if (if (local-variable-p 'forge-buffer-draft-p)
+                            forge-buffer-draft-p
+                          .draft)
+                        (concat "Draft: " .title)
+                      .title))
+          (description . , .body)
+          ,@(and (not (equal head-remote base-remote))
+                 `((target_project_id . ,(oref base-repo forge-id))))
+          (target_branch . ,base-branch)
+          (source_branch . ,head-branch)
+          (allow_collaboration . t))
+        :callback  (forge--post-submit-callback)
+        :errorback (forge--post-submit-errorback)))))
+
 (cl-defmethod forge--submit-create-post ((_ forge-gitlab-repository) topic)
   (forge--glab-post topic
     (if (forge-issue-p topic)
@@ -490,23 +515,44 @@
       (forge-pullreq "/projects/:project/merge_requests/:number")
       (forge-issue   "/projects/:project/issues/:number"))
     `((,field . ,value))
-    :callback (forge--set-field-callback)))
+    :callback (forge--set-field-callback topic)))
 
 (cl-defmethod forge--set-topic-title
   ((repo forge-gitlab-repository) topic title)
   (forge--set-topic-field repo topic 'title title))
 
 (cl-defmethod forge--set-topic-state
-  ((repo forge-gitlab-repository) topic)
+  ((repo forge-gitlab-repository) topic state)
   (forge--set-topic-field repo topic 'state_event
-                          (cl-ecase (oref topic state)
-                            (closed "reopen")
-                            (open   "close"))))
+                          (pcase-exhaustive state
+                            ;; Merging isn't done through here.
+                            ('completed "close")
+                            ('unplanned "close")
+                            ('rejected  "close")
+                            ('open      "reopen"))))
+
+(cl-defmethod forge--set-topic-draft
+  ((repo forge-gitlab-repository) topic value)
+  (let ((buffer (current-buffer)))
+    (glab-graphql
+     `(mutation (mergeRequestSetDraft
+                 [(input $input MergeRequestSetDraftInput!)]
+                 (mergeRequest iid draft)))
+     `((input (projectPath . ,(oref repo slug))
+              (iid . ,(number-to-string (oref topic number)))
+              (draft . ,value)))
+     :host (oref (forge-get-repository topic) apihost)
+     :auth 'forge
+     :callback (lambda (data &rest _)
+                 (if (assq 'error data)
+                     (ghub--graphql-pp-response data)
+                   (oset topic draft-p value)
+                   (forge-refresh-buffer buffer))))))
 
 (cl-defmethod forge--set-topic-labels
   ((repo forge-gitlab-repository) topic labels)
   (forge--set-topic-field repo topic 'labels
-                          (mapconcat #'identity labels ",")))
+                          (string-join labels ",")))
 
 (cl-defmethod forge--set-topic-assignees
   ((repo forge-gitlab-repository) topic assignees)
@@ -514,10 +560,19 @@
     (cl-typecase topic
       (forge-pullreq ; Can only be assigned to a single user.
        (forge--set-topic-field repo topic 'assignee_id
-                               (caddr (assoc (car assignees) users))))
+                               (or (caddr (assoc (car assignees) users))
+                                   0)))
       (forge-issue
        (forge--set-topic-field repo topic 'assignee_ids
-                               (--map (caddr (assoc it users)) assignees))))))
+                               (or (--map (caddr (assoc it users)) assignees)
+                                   0))))))
+
+(cl-defmethod forge--set-topic-review-requests
+  ((repo forge-gitlab-repository) topic reviewers)
+  (let ((users (mapcar #'cdr (oref repo assignees))))
+    (forge--set-topic-field repo topic 'reviewer_ids
+                            (or (--map (caddr (assoc it users)) reviewers)
+                                0))))
 
 (cl-defmethod forge--delete-comment
   ((_repo forge-gitlab-repository) post)
@@ -528,15 +583,15 @@
       (forge-issue-post
        "/projects/:project/issues/:topic/notes/:number")))
   (closql-delete post)
-  (magit-refresh))
+  (forge-refresh-buffer))
 
-(cl-defmethod forge--topic-templates ((repo forge-gitlab-repository)
-                                      (_ (subclass forge-issue)))
+(cl-defmethod forge--topic-template-files ((repo forge-gitlab-repository)
+                                           (_ (subclass forge-issue)))
   (--filter (string-match-p "\\`\\.gitlab/issue_templates/.+\\.md\\'" it)
             (magit-revision-files (oref repo default-branch))))
 
-(cl-defmethod forge--topic-templates ((repo forge-gitlab-repository)
-                                      (_ (subclass forge-pullreq)))
+(cl-defmethod forge--topic-template-files ((repo forge-gitlab-repository)
+                                           (_ (subclass forge-pullreq)))
   (--filter (string-match-p "\\`\\.gitlab/merge_request_templates/.+\\.md\\'" it)
             (magit-revision-files (oref repo default-branch))))
 
@@ -553,13 +608,10 @@
                                     topic hash method)
   (forge--glab-put topic
     "/projects/:project/merge_requests/:number/merge"
-    `((squash . ,(if (eq method 'squash) "true" "false"))
+    `((squash . ,(eq method 'squash))
       ,@(and hash `((sha . ,hash))))))
 
-;;; Utilities
-
-(cl-defmethod forge--topic-type-prefix ((_repo forge-gitlab-repository) type)
-  (if (eq type 'pullreq) "!" "#"))
+;;; Wrappers
 
 (cl-defun forge--glab-get (obj resource
                                &optional params
